#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "litefs.h"
#include "litefs.p"

static void usage()
{
	fprintf (stderr,"litefs-proxy version 1.0\n"
		"\n"
		"litefs-proxy --socket sock_path [ --fork ] [ --debug] [ --exec command ...]\n"
		"litefs-proxy --pipe [ --fork] [ --debug] [ --exec command ...]\n");
}

static bool debug=false;

static void n_debug (const char *ctl, ...)
{
	if (debug){
		va_list list;
		va_start (list,ctl);
		vfprintf (stderr,ctl,list);
		va_end (list);
	}
}


static int proxy_read (
	int in,
	char *buf,
	int size)
{
	// We always assume a single packet, which is wrong
	// We will use the len field of the LITE_REQUEST field
	return read (in,buf,size);
}

static int proxy_doreq (
	int client_in,
	int client_out,
	int proxy_in,
	int proxy_out)
{
	int ret = -1;
	char buf[10000];
	int n = proxy_read (client_in,buf,sizeof(buf)-1);
	if (n > 0){
		n_debug ("-proxy write %d\n",n);
		if (write (proxy_out,buf,n) == n){
			n = proxy_read (proxy_in,buf,sizeof(buf)-1);
			n_debug ("-proxy read %d\n",n);
			if (n > 0){
				if (write (client_out,buf,n)==n){
					ret = 0;
				}
			}
		}
	}
	return ret;
}



static int proxy_fdset (fd_set &fdin, int maxfd, int handle)
{
	if (handle != -1){
		FD_SET (handle,&fdin);
		if (handle > maxfd) maxfd = handle;
	}
	return maxfd;
}

static void proxy_loop (
	int sock,
	int pipein,
	int pipeout,
	int proxyin,
	int proxyout)
{
	int tbfd[1024];
	int nbfd = 0;
	fd_set fdin;
	FD_ZERO (&fdin);
	int maxfd = 0;
	maxfd = proxy_fdset (fdin,maxfd,sock);
	maxfd = proxy_fdset (fdin,maxfd,pipein);
	maxfd = proxy_fdset (fdin,maxfd,pipeout);
	while (1){
		fd_set in;
		memcpy (&in,&fdin,sizeof(in));
		n_debug ("select maxfd %d\n",maxfd);
		if (select (maxfd+1,&in,NULL,NULL,NULL)!=-1){
			int orig_nbfd = nbfd;
			if (sock != -1 && FD_ISSET (sock,&in)){
				struct sockaddr_un unc;
				size_t len = sizeof(unc);
				unc.sun_family = AF_UNIX;
				int s = accept (sock,(struct sockaddr*)&unc,&len);
				tbfd[nbfd++] = s;
				FD_SET (s,&fdin);
				if (s >= maxfd) maxfd = s + 1;
			}
			if (pipein != -1 && FD_ISSET(pipein,&in)){
				if (proxy_doreq (pipein,pipeout,proxyin,proxyout) == -1){
					break;
				}
			}
			int *ptfd = tbfd;
			bool redo_fdin = false;
			for (int i=0; i< orig_nbfd; i++, ptfd++){
				int fd = *ptfd;
				if (FD_ISSET(fd,&in)){
					if (proxy_doreq (fd,fd,proxyin,proxyout) == -1){
						close (fd);
						redo_fdin = true;
						*ptfd = -1;
					}
				}
			}
			if (redo_fdin){
				FD_ZERO (&fdin);
				maxfd = 0;
				maxfd = proxy_fdset (fdin,maxfd,sock);
				maxfd = proxy_fdset (fdin,maxfd,pipein);
				ptfd = tbfd;
				int *copyfd = tbfd;
				for (int i=0; i<nbfd; i++, ptfd++){
					int fd = *ptfd;
					if (fd != -1){
						maxfd = proxy_fdset (fdin,maxfd,fd);
						*copyfd++ = fd;
					}
				}
				nbfd = (int)(copyfd - tbfd);
			}
		}
	}
}

int main (int argc, char *argv[])
{
	int ret = -1;
	if (argc == 1){
		usage();
	}else{
		int sock = -1;
		int pipein = -1;
		int pipeout = -1;
		int proxyin = -1;
		int proxyout = -1;
		bool dofork = false;
		for (int i=1; i<argc; i++){
			if (strcmp(argv[i],"--socket")==0){
				i++;
				sock = litefsd_server(argv[i]);
			}else if (strcmp(argv[i],"--debug")==0){
				debug = true;
			}else if (strcmp(argv[i],"--fork")==0){
				dofork = true;
			}else if (strcmp(argv[i],"--pipe")==0){
				pipein = 0;
				pipeout = 1;
			}else if (strcmp(argv[i],"--exec")==0){
				if (i == argc -1){
					usage();
					exit (-1);
				}else{
					int tbin[2],tbout[2];
					if (pipe(tbin)==-1 || pipe(tbout)==-1){
						fprintf (stderr,"Can't setup pipe (%s)\n",strerror(errno));
					}else{
						int pid = fork();
						if (pid == -1){
							fprintf (stderr,"Can't fork new process (%s)\n"
								,strerror(errno));
						}else if (pid == 0){
							close (tbout[1]);
							close (tbin[0]);
							dup2 (tbin[1],1);
							dup2 (tbout[0],0);
							close (tbout[0]);
							close (tbin[1]);
							char line[2000];
							line[0] = '\0';
							for (int j=i+1; j<argc; j++){
								strcat (line,argv[j]);
								strcat (line," ");
							}
							system (line);
							_exit (0);
						}else{
							close (tbout[0]);
							close (tbin[1]);
							proxyin = tbin[0];
							proxyout = tbout[1];
						}
						i=argc-1;
					}
				}
			}else{
				usage();
				exit (-1);
			}
		}
		if ((sock != -1 || pipein != -1) && proxyin != -1){
			if (dofork && fork() != 0) exit(0);
			proxy_loop(sock,pipein,pipeout,proxyin,proxyout);
		}else{
			usage();
		}
	}
	return ret;
}




